"""Utility functions for creating an interpreter."""

from collections import OrderedDict
from copy import deepcopy
from io import StringIO
import os
import re
from functools import partial
from tempfile import TemporaryDirectory
from typing import List

from IPython.core.display import display

from robot.api import get_model
from robot.errors import DataError
from robot.reporting import ResultWriter
from robot.running.model import TestSuite
from robot.running.builder.testsettings import TestDefaults
from robot.running.builder.parsers import ErrorReporter
from robot.running.builder.transformers import SettingsBuilder, SuiteBuilder
from robot.model.itemlist import ItemList

from ipywidgets import VBox, HBox, Button, Output, Text

from .utils import (
    detect_robot_context, line_at_cursor, scored_results,
    complete_libraries, get_lunr_completions, remove_prefix,
    display_log, process_screenshots, lunr_query, get_keyword_doc
)
from .selectors import (
    BrokenOpenConnection, clear_selector_highlights, get_autoit_selector_completions, get_selector_completions,
    get_white_selector_completions, get_win32_selector_completions, is_autoit_selector,
    is_selector, is_white_selector, is_win32_selector, close_current_connection, yield_current_connection
)
from .constants import VARIABLE_REGEXP, BUILTIN_VARIABLES
from .listeners import RobotKeywordsIndexerListener

from robot.running.model import UserKeyword
UserKeyword.source = property(lambda self: self.actual_source)


def normalize_argument(name):
    if "=" in name:
        name, default = name.split("=", 1)
    else:
        default = None

    return (
        name,
        re.sub(r"\W", "_", re.sub(r"^[^\w]*|[^\w]*$", "", name, re.U), re.U),
        default
    )


def execute_keyword(suite: TestSuite, name, arguments, **values):
    header = suite.rpa and "Tasks" or "Test Cases"
    code = f"""\

*** {header} ***

{name}
    {name}  {'  '.join([values[a[1]] for a in arguments])}
"""

    # Copy the test suite
    suite = deepcopy(suite)
    suite.rpa = True

    with TemporaryDirectory() as path:
        _, report = _execute_impl(code, suite, outputdir=path, interactive_keywords=False)

    if report is not None:
        display(report, raw=True)


def on_button_execute(execute, controls, out, widgets, *args, **kwargs):
    values = {key: control.value for key, control in controls.items()}

    with out:
        description = widgets[0].description
        widgets[0].description = "Executing..."

        for widget in widgets:
            widget.disabled = True

        out.clear_output(wait=True)

        try:
            execute(**values)
        finally:
            widgets[0].description = description
            for widget in widgets:
                widget.disabled = False


def get_interactive_keyword(suite: TestSuite, keyword):
    """Get an interactive widget for testing a keyword."""
    name = keyword.name
    arguments = [normalize_argument(arg) for arg in keyword.args]

    # Make a copy of the suite, the suite the widget operates on must no be
    # the same as the main one
    suite_copy = deepcopy(suite)

    execute_key = partial(execute_keyword, suite_copy, name, arguments)

    widgets = []
    controls = OrderedDict()
    out = Output()

    for arg in arguments:
        input_widget = Text(description=arg[1] + "=", value=arg[2])
        widgets.append(input_widget)
        controls[arg[1]] = input_widget

    button = Button(description=name)
    button.on_click(partial(on_button_execute, execute_key, controls, out, widgets))
    widgets.insert(0, button)

    return VBox((HBox(widgets), out))


class TestSuiteError(Exception):
    pass


class ProgressUpdater(StringIO):
    """Wrapper designed to capture robot.api.logger.console and display it.
    This can be used passing an instance of this to the execute's stdout argument"""

    colors = re.compile(r"\[[0-?]+[^m]+m")

    def __init__(self, display, update_display):
        self.display = display
        self.update_display = update_display

        self.progress = {"test": "n/a", "keyword": "n/a", "message": None}
        self.already_displayed = False

        super(ProgressUpdater, self).__init__()

    def _update(self):
        status_line = " | ".join(
            str(s)
            for s in [
                self.progress["test"],
                self.progress["keyword"],
                self.progress["message"],
            ]
            if s
        )

        mimebundle = {
            "text/html": f'<pre style="white-space:nowrap;overflow:hidden;padding-left:1ex;'
                         f'"><i class="fa fa-spinner fa-pulse"></i>{status_line}</pre>'
        }

        if not self.already_displayed:
            self.display(mimebundle)
            self.already_displayed = True
        else:
            self.update_display(mimebundle)

    def update(self, data):
        if "test" in data:
            self.progress["test"] = data["test"]
            self.progress["message"] = None
        elif "keyword" in data:
            self.progress["keyword"] = data["keyword"]
            self.progress["message"] = None
        self._update()

    def clear(self):
        self.update_display({"text/plain": ""})

    def write(self, s):
        self.progress["message"] = s.strip()
        self._update()
        return super(ProgressUpdater, self).write(s)


class ErrorStream():
    def __init__(self):
        self.message = ''

    def write(self, message, flush=False):
        self.message = self.message + message

        if flush:
            self.flush()

    def flush(self):
        message_copy = str(self.message)

        self.message = ''

        raise TestSuiteError(message_copy)


class NoOpStream():
    def write(self, message, flush=False):
        # This is a no-op
        pass

    def flush(self):
        # This is a no-op
        pass


def init_suite(name: str, source: str = os.getcwd()):
    """Create a new test suite."""
    return TestSuite(name=name, source=source)


def generate_report(suite: TestSuite, outputdir: str):
    process_screenshots(outputdir)

    writer = ResultWriter(os.path.join(outputdir, "output.xml"))
    writer.write_results(
        log=os.path.join(outputdir, "log.html"),
        report=None,
        rpa=getattr(suite, "rpa", False),
    )

    with open(os.path.join(outputdir, "log.html"), "rb") as fp:
        log = fp.read()
        log = log.replace(b'"reportURL":"report.html"', b'"reportURL":null')

    html = """
        <button
          class="jp-mod-styled jp-mod-accept"
          onClick="{};event.preventDefault();event.stopPropagation();"
        >
            <i class="fa fa-file" aria-hidden="true"></i>
            Log
        </button>
        """.format(display_log(log, "log.html"))

    return {"text/html": html}


def _execute_impl(code: str, suite: TestSuite, defaults: TestDefaults = TestDefaults(),
                  stdout=None, stderr=None, listeners=[], drivers=[], outputdir=None, interactive_keywords=True):
    # Clear selector completion highlights
    for driver in yield_current_connection(drivers, ["RPA.Browser", "selenium", "jupyter"]):
        try:
            clear_selector_highlights(driver)
        except BrokenOpenConnection:
            close_current_connection(drivers, driver)

    # Copy keywords/variables/libraries in case of failure
    imports = get_items_copy(suite.resource.imports)
    variables = get_items_copy(suite.resource.variables)
    keywords = get_items_copy(suite.resource.keywords)

    # Compile AST
    model = get_model(
        StringIO(code),
        data_only=False,
        curdir=os.getcwd().replace("\\", "\\\\"),
    )
    ErrorReporter(code).visit(model)
    SettingsBuilder(suite, defaults).visit(model)
    SuiteBuilder(suite, defaults).visit(model)

    # Strip variables/keyword duplicates
    strip_duplicate_items(suite.resource.variables)
    strip_duplicate_items(suite.resource.keywords)

    new_imports = [item for item in get_items_copy(suite.resource.imports) if item not in imports]
    for new_import in new_imports:
        new_import.source = suite.source
    new_variables = [item for item in get_items_copy(suite.resource.variables) if item not in variables]
    for new_variable in new_variables:
        new_variable.source = suite.source
    # If there is no test, allow the user to interact with defined keywords by providing widgets
    new_keywords = [item for item in get_items_copy(suite.resource.keywords) if item not in keywords]
    for new_keyword in new_keywords:
        new_keyword.actual_source = suite.source
    if not suite.tests and new_keywords and interactive_keywords:
        return None, [get_interactive_keyword(suite, keyword) for keyword in new_keywords]

    # Set default streams
    # By default stdout is no-op
    # By default stderr raises an exception when flushing (workaround robotframework which does not raise)
    if stdout is None:
        stdout = NoOpStream()
    if stderr is None:
        stderr = ErrorStream()

    # Execute suite
    try:
        result = suite.run(outputdir=outputdir, stdout=stdout, stderr=stderr, listener=listeners)
    except TestSuiteError as e:
        # Reset keywords/variables/libraries
        set_items(suite.resource.imports, imports)
        set_items(suite.resource.variables, variables)
        set_items(suite.resource.keywords, keywords)

        clean_items(suite.tests)

        raise e

    for listener in listeners:
        if isinstance(listener, RobotKeywordsIndexerListener):
            listener.import_from_suite_data(suite)

    # Detect RPA
    suite.rpa = get_rpa_mode(model)

    report = None
    if suite.tests:
        report = generate_report(suite, outputdir)

    # Remove tests run so far,
    # this is needed so that we don't run them again in the next execution
    clean_items(suite.tests)

    return result, report


def execute(code: str, suite: TestSuite, defaults: TestDefaults = TestDefaults(),
            stdout=None, stderr=None, listeners=[], drivers=[], outputdir=None):
    """
    Execute a snippet of code, given the current test suite. Returns a tuple containing the result of the
    suite (if there were tests) and a displayable object containing either the report or interactive widgets.
    """
    if outputdir is None:
        with TemporaryDirectory() as path:
            result = _execute_impl(code, suite, defaults, stdout, stderr, listeners, drivers, path)
    else:
        result = _execute_impl(code, suite, defaults, stdout, stderr, listeners, drivers, outputdir)

    return result


def complete(code: str, cursor_pos: int, suite: TestSuite, keywords_listener: RobotKeywordsIndexerListener = None, extra_libraries: List[str] = [], drivers=[]):
    """Complete a snippet of code, given the current test suite."""
    context = detect_robot_context(code, cursor_pos)
    cursor_pos = cursor_pos is None and len(code) or cursor_pos
    line, offset = line_at_cursor(code, cursor_pos)
    line_cursor = cursor_pos - offset
    needle = re.split(r"\s{2,}|\t| \| ", line[:line_cursor])[-1].lstrip()

    library_completion = context == "__settings__" and any(
        [
            line.lower().startswith("library "),
            "import library " in line.lower(),
            "reload library " in line.lower(),
            "get library instance" in line.lower(),
        ]
    )

    matches = []

    # Try to complete a variable
    if needle and needle[0] in "$@&%":
        potential_vars = list(set(
            [var.name for var in suite.resource.variables] +
            VARIABLE_REGEXP.findall(code) +
            BUILTIN_VARIABLES
        ))

        matches = [
            m["ref"]
            for m in scored_results(needle, [dict(ref=v) for v in potential_vars])
            if needle.lower() in m["ref"].lower()
        ]

        if len(line) > line_cursor and line[line_cursor] == "}":
            cursor_pos += 1
            needle += "}"
    # Try to complete a library name
    elif library_completion:
        needle = needle.lower()
        needle = remove_prefix(needle, 'library ')
        needle = remove_prefix(needle, 'import library ')
        needle = remove_prefix(needle, 'reload library ')
        needle = remove_prefix(needle, 'get library instance ')

        matches = complete_libraries(needle, extra_libraries)
    # Try to complete a CSS selector
    elif is_selector(needle):
        matches = []
        for driver in yield_current_connection(drivers, ["RPA.Browser", "selenium", "jupyter", "appium"]):
            matches = [get_selector_completions(needle.rstrip(), driver)[0]]
    # Try to complete an AutoIt selector
    elif is_autoit_selector(needle):
        matches = [get_autoit_selector_completions(needle)[0]]
    # Try to complete a white selector
    elif is_white_selector(needle):
        matches = [get_white_selector_completions(needle)[0]]
    # Try to complete a Windows selector
    elif is_win32_selector(needle):
        matches = [get_win32_selector_completions(needle)[0]]
    # Try to complete a keyword
    elif keywords_listener is not None:
        matches = get_lunr_completions(
            needle,
            keywords_listener.index,
            keywords_listener.keywords,
            context
        )

    return {
        "matches": matches,
        "cursor_end": cursor_pos,
        "cursor_start": cursor_pos - len(needle)
    }


def inspect(code: str, cursor_pos: int, suite: TestSuite, keywords_listener: RobotKeywordsIndexerListener = None, detail_level=0):
    cursor_pos = len(code) if cursor_pos is None else cursor_pos
    line, offset = line_at_cursor(code, cursor_pos)
    line_cursor = cursor_pos - offset
    left_needle = re.split(r"\s{2,}|\t| \| ", line[:line_cursor])[-1]
    right_needle = re.split(r"\s{2,}|\t| \| ", line[line_cursor:])[0]
    needle = left_needle.lstrip().lower() + right_needle.rstrip().lower()

    results = []
    data = {}
    found = False

    if needle and lunr_query(needle):
        query = lunr_query(needle)
        results = keywords_listener.index.search(query)
        results += keywords_listener.index.search(query.strip("*"))

    for result in results:
        keyword = keywords_listener.keywords[result["ref"]]

        if needle not in [keyword.name.lower(), result["ref"].lower()]:
            continue

        data = get_keyword_doc(keyword)
        found = True
        break

    return {
        "data": data,
        "found": found,
    }


def shutdown_drivers(drivers=[]):
    for driver in drivers:
        if hasattr(driver["instance"], "quit"):
            driver["instance"].quit()


def strip_duplicate_items(items: ItemList):
    """Remove duplicates from an item list."""
    new_items = {}
    for item in items:
        new_items[item.name] = item
    items._items = list(new_items.values())


def clean_items(items: ItemList):
    """Remove elements from an item list."""
    items._items = []


def set_items(items: ItemList, value: List):
    """Remove elements from an item list."""
    items._items = value


def get_items_copy(items: ItemList):
    """Get copy of an itemlist."""
    return list(items._items)


def get_rpa_mode(model):
    """Get RPA mode for the test suite."""
    if not model:
        return None
    tasks = [s.tasks for s in model.sections if hasattr(s, 'tasks')]
    if all(tasks) or not any(tasks):
        return tasks[0] if tasks else None
    raise DataError('One file cannot have both tests and tasks.')



from robot.libdocpkg.model import KeywordDoc
import re

VARIABLE_REGEXP = re.compile(r"[$@&%]\{[\w\s]+\}")

BUILTIN_VARIABLES = [
    "${TEMPDIR}",
    "${EXECDIR}",
    "${/}",
    "${:}",
    "${\\n}",
    "${SPACE}",
    "${True}",
    "${False}",
    "${None}",
    "${null}",
    "${OUTPUT_DIR}",
    "${OUTPUT_FILE}",
    "${REPORT_FILE}",
    "${LOG_FILE}",
    "${DEBUG_FILE}",
    "${LOG_LEVEL}",
    "${PREV_TEST_NAME}",
    "${PREV_TEST_STATUS}",
    "${PREV_TEST_MESSAGE}",
]

NAME_REGEXP = re.compile("`(.+?)`")

CONTEXT_LIBRARIES = {
    "__root__": list(
        map(
            lambda item: KeywordDoc(name=item[0], doc=item[1]),
            {
                "*** Settings ***": """\
The Setting table is used to import test libraries, resource files and variable
files and to define metadata for test suites and test cases. It can be included
in test case files and resource files. Note that in a resource file, a Setting
table can only include settings for importing libraries, resources, and
variables.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#appendices
""",  # noqa: E501
                "*** Variables ***": """\
The most common source for variables are Variable tables in test case files and
resource files. Variable tables are convenient, because they allow creating
variables in the same place as the rest of the test data, and the needed syntax
is very simple. Their main disadvantages are that values are always strings and
they cannot be created dynamically. If either of these is a problem, variable
files can be used instead.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#creating-variables
""",  # noqa: E501
                "*** Test Cases ***": """\
Test cases are constructed in test case tables from the available keywords.
Keywords can be imported from test libraries or resource files, or created in
the keyword table of the test case file itself.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-case-syntax
""",  # noqa: E501
                "*** Tasks ***": """\
Tasks are constructed in tasks tables from the available keywords. Keywords can
be imported from test libraries or resource files, or created in the keyword
table of the tasks file itself.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-case-syntax
""",  # noqa: E501
                "*** Keywords ***": """\
User keyword names are in the first column similarly as test cases names. Also
user keywords are created from keywords, either from keywords in test libraries
or other user keywords. Keyword names are normally in the second column, but
when setting variables from keyword return values, they are in the subsequent
columns.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#user-keyword-syntax
""",  # noqa: E501
            }.items(),
        )
    ),
    "__settings__": list(
        map(
            lambda item: KeywordDoc(name=item[0], doc=item[1]),
            {
                "Library": """\
Test libraries are typically imported using the Library setting.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#importing-libraries
""",  # noqa: E501
                "Resource": """\
Resource files are imported using the Resource setting in the Settings table.
The path to the resource file is given in the cell after the setting name.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#taking-resource-files-into-use
""",  # noqa: E501
                "Variables": """\
All test data files can import variables using the Variables setting in the
Setting table, in the same way as resource files are imported using the
Resource setting.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#taking-variable-files-into-use
""",  # noqa: E501
                "Documentation": """\
Used for specifying a test suite or resource file documentation.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-suite-name-and-documentation
""",  # noqa: E501
                "Metadata": """\
Used for setting free test suite metadata.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#free-test-suite-metadata
""",  # noqa: E501
                "Suite Setup": """\
Used for specifying the suite setup.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#suite-setup-and-teardown
""",  # noqa: E501
                "Suite Teardown": """\
Used for specifying the suite teardown.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#suite-setup-and-teardown
""",  # noqa: E501
                "Test Setup": """\
Used for specifying a default test setup.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-setup-and-teardown
""",  # noqa: E501
                "Test Teardown": """\
Used for specifying a default test teardown.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-setup-and-teardown
""",  # noqa: E501
                "Test Template": """\
Used for specifying a default template keyword for test cases.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-templates
""",  # noqa: E501
                "Test Timeout": """\
Used for specifying a default test case timeout.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-case-timeout
""",  # noqa: E501
                "Task Setup": """\
Used for specifying a default task setup.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-setup-and-teardown
""",  # noqa: E501
                "Task Teardown": """\
Used for specifying a default task teardown.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-setup-and-teardown
""",  # noqa: E501
                "Task Template": """\
Used for specifying a default template keyword for tasks.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-templates
""",  # noqa: E501
                "Task Timeout": """\
Used for specifying a default task timeout.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-case-timeout
""",  # noqa: E501
                "Force Tags": """\
Used for specifying forced values for tags when tagging test cases.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#tagging-test-cases
""",  # noqa: E501
                "Default Tags": """\
Used for specifying default values for tags when tagging test cases.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#tagging-test-cases
""",  # noqa: E501
            }.items(),
        )
    ),
    "__tasks__": list(
        map(
            lambda item: KeywordDoc(name=item[0], doc=item[1]),
            {
                "[Documentation]": """\
Used for specifying test case documentation.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-case-name-and-documentation
""",  # noqa: E501
                "[Tags]": """\
Used for tagging test cases.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#tagging-test-cases
""",  # noqa: E501
                "[Setup]": """\
Used for specifying a test setup.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-setup-and-teardown
""",  # noqa: E501
                "[Teardown]": """\
Used for specifying a test teardown.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-setup-and-teardown
""",  # noqa: E501
                "[Template]": """\
Used for specifying a template keyword.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-templates
""",  # noqa: E501
                "[Timeout]": """\
Used for specifying a test case timeout.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#test-case-timeout
""",  # noqa: E501
            }.items(),
        )
    ),
    "__keywords__": list(
        map(
            lambda item: KeywordDoc(name=item[0], doc=item[1]),
            {
                "[Documentation]": """\
Used for specifying a user keyword documentation.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#user-keyword-documentation
""",  # noqa: E501
                "[Tags]": """\
Used for specifying user keyword tags.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#user-keyword-tags
""",  # noqa: E501
                "[Arguments]": """\
Used for specifying user keyword arguments.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#user-keyword-arguments
""",  # noqa: E501
                "[Return]": """\
Used for specifying user keyword return values.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#user-keyword-return-values
""",  # noqa: E501
                "[Teardown]": """\
Used for specifying user keyword teardown.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#user-keyword-teardown
""",  # noqa: E501
                "[Timeout]": """\
Used for specifying a user keyword timeout.

`Robot Framework User Guide`__

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#user-keyword-timeout
""",  # noqa: E501
            }.items(),
        )
    ),
}

CONTEXT_LIBRARIES["__settings__"].extend(CONTEXT_LIBRARIES["__root__"])


SCRIPT_DISPLAY_LOG = """\
var content = '{content}';

var w = window.open('', '', 'width=900,height=900');
w.document.body.style = 'margin:0; overflow: hidden;';

var i = w.document.createElement('iframe');
i.src = 'data:text/html;base64,' + content;
i.style = (
  'position:absolute;' +
  'left: 0px; width: 100%;' +
  'top: 0px; height: 100%;' +
  'border-width: 0;'
);
w.document.body.append(i);

var a = w.document.createElement('a');
a.appendChild(w.document.createTextNode('Download'));
a.href = 'data:text/html;base64,' + content;
a.download = '{filename}';
a.style = (
  'position:fixed;top:0;right:0;' +
  'color:white;background:black;text-decoration:none;' +
  'font-weight:bold;padding:7px 14px;border-radius:0 0 0 5px;'
);
w.document.body.append(a);
"""



import os
import pkg_resources
import re
import time

try:
    import WhiteLibrary
    from .selectors_white import PickSnipTool
except ImportError:
    WhiteLibrary = None
    PickSnipTool = None

try:
    import pywinauto
    from .selectors_pywin32 import ImageRegionTool, PointPickTool
except ImportError:
    ImageRegionTool = None
    PointPickTool = None

try:
    import pywintypes
    import win32com.client

    try:
        AutoIt = win32com.client.Dispatch("AutoItX3.Control")
    except pywintypes.com_error:
        AutoIt = None
except ImportError:
    AutoIt = None

try:
    from selenium.common.exceptions import WebDriverException
    from selenium.common.exceptions import InvalidSessionIdException
    from selenium.common.exceptions import TimeoutException
except ImportError:

    class WebDriverException(Exception):
        pass

    class TimeoutException(Exception):
        pass


class BrokenOpenConnection(Exception):
    """Broken sticky connection that should be closed."""

    def __init__(self, connection):
        """Init with connection be closed."""
        self.connection = connection


def yield_current_connection(connections, types_):
    current_connections = [
        connection["instance"]
        for connection in connections
        if connection["type"] in types_ and connection["current"]
    ]

    for instance in current_connections:
        yield instance
        break


def close_current_connection(connections, connection_to_close):
    match = None
    for connection in connections:
        if connection["instance"] is connection_to_close:
            match = connection
            break
    if match is not None:
        if hasattr(match["instance"], "quit"):
            match["instance"].quit()
        connections.remove(match)


IS_SELENIUM_SELECTOR_NEEDLE = re.compile(
    r"^id=|^id:|"
    r"^name=|^name:|"
    r"^css=|^css:|"
    r"^tag=|^tag:|"
    r"^link=|^link:|"
    r"^partial link=|^partial link:|"
    r"^xpath=|^xpath:"
)
IS_APPIUM_SELECTOR_NEEDLE = re.compile(r"^id=|^id:|^xpath=|^xpath:")
IS_AUTOIT_SELECTOR_NEEDLE = re.compile(r"^strTitle=|^strControl=")
IS_WHITE_SELECTOR_NEEDLE = re.compile(r"^ae=|^ae:|^template=|^template:")
IS_WIN32_SELECTOR_NEEDLE = re.compile(r"^uia=|^uia:|^template=|^template:")
IS_ID_SELECTOR_NEEDLE = re.compile(r"^id=|^id:")
IS_NAME_SELECTOR_NEEDLE = re.compile(r"^name=|^name:")
IS_CSS_SELECTOR_NEEDLE = re.compile(r"^css=|^css:")
IS_TAG_SELECTOR_NEEDLE = re.compile(r"^tag=|^tag:")
IS_LINK_SELECTOR_NEEDLE = re.compile(r"^link=|^link:|" r"^partial link:|^partial link=")
IS_XPATH_SELECTOR_NEEDLE = re.compile(r"^xpath=|^xpath:")
FORM_TAG_NAMES = ["input", "textarea", "select", "button", "datalist"]
IS_TEXT = re.compile(r"^[\w\s]+$", re.U)

SELECTOR_HIGHLIGHT_STYLE_SCRIPT = """
(function() {
  var node = document.createElement('style');
  node.setAttribute('data-name', 'robotframework-interpreter');
  node.innerHTML = '' +
    '[data-robotframework-interpreter] {' +
      'outline: 2px solid red !important;' +
      'opacity: 1.0 !important;' +
    '}' +
    '#robotframework-interpreter-picker { ' +
      'position: fixed;' +
      'top: 0;' +
      'right: 0;' +
      'bottom: 0;' +
      'left: 0;' +
      'z-index: 9999;' +
      'cursor: crosshair;' +
    '}' +
    '#robotframework-interpreter-picker::before { ' +
      'display: block;' +
      'content: "";' +
      'position: absolute;' +
      'top: 0;' +
      'right: 0;' +
      'bottom: 0;' +
      'left: 0;' +
      'background: black;' +
      'opacity: 0.25;' +
      'z-index: -1;' +
    '}' +
    '#robotframework-interpreter-picker::after { ' +
      'display: block;' +
      'content: "Click element to select it... (before Selenium timeout).";' +
      'font-size: 12px;' +
      'text-align: center;' +
      'top: 3px;' +
      'margin-right: auto;' +
      'width: 50%;' +
      'margin-left: auto;' +
      'padding: 3px;' +
      'background: white;' +
      'opacity: 1.0;' +
      'border: 1px solid black;' +
    '}';
  document.head.appendChild(node);
})();
"""


def is_selenium_selector(needle):
    return bool(IS_SELENIUM_SELECTOR_NEEDLE.match(needle))


def is_appium_selector(needle):
    return bool(IS_APPIUM_SELECTOR_NEEDLE.match(needle))


def is_autoit_selector(needle):
    return bool(IS_AUTOIT_SELECTOR_NEEDLE.match(needle))


def is_white_selector(needle):
    if PickSnipTool is None:
        return False
    else:
        return bool(IS_WHITE_SELECTOR_NEEDLE.match(needle))


def is_win32_selector(needle):
    if PointPickTool is None or ImageRegionTool is None:
        return False
    else:
        return bool(IS_WIN32_SELECTOR_NEEDLE.match(needle))


def is_selector(needle):
    return is_selenium_selector(needle) or is_appium_selector(needle)


def get_element_highlight_script(results, old_elements):
    script = ""
    counter = 0
    arguments = []
    elements = [r[1] for r in results]
    for element in old_elements:
        if element not in elements:
            script += f'arguments[{counter:d}].removeAttribute("data-robotframework-interpreter");'
            arguments.append(element)
            counter += 1
    for completion, element in results:
        if element not in old_elements:
            script += (
                f"arguments[{counter:d}].setAttribute("
                f'"data-robotframework-interpreter", "{completion}");'
            )
            arguments.append(element)
            counter += 1
    return script, arguments


def clear_selector_highlights(driver):
    try:
        script, arguments = get_element_highlight_script(
            [], driver.find_elements_by_css_selector("[data-robotframework-interpreter]")
        )
    except InvalidSessionIdException:
        raise BrokenOpenConnection(driver)
    except WebDriverException:
        return
    if script:
        driver.execute_script(script, *arguments)


def get_selector_completions(needle, driver):
    if repr(driver).startswith("<appium.webdriver"):
        return get_appium_selector_completions(needle, driver)
    else:
        return get_selenium_selector_completions(needle, driver)


def get_selenium_selector_completions(needle, driver):
    try:
        # Inject supporting JS and CSS
        styles = 'style[data-name="robotframework-interpreter"]'
        if not driver.find_elements_by_css_selector(styles):
            with pkg_resources.resource_stream(
                "robotframework_interpreter", "resources/simmerjs/simmer.js"
            ) as fp:
                driver.execute_script(
                    fp.read().decode("utf-8") + SELECTOR_HIGHLIGHT_STYLE_SCRIPT
                )
        can_highlight = True
    except WebDriverException:
        can_highlight = False

    # Get results
    try:
        results = _get_selenium_selector_completions(needle, driver)
    except WebDriverException as e:
        return ["Exception (press esc to clear):", str(e)]

    # Highlight
    if can_highlight:
        script, arguments = get_element_highlight_script(
            results, driver.find_elements_by_css_selector("[data-robotframework-interpreter]")
        )
        driver.execute_script(script, *arguments)

    # Return
    return [r[0] for r in results]


def get_appium_selector_completions(needle, driver):
    # Get results
    try:
        results = _get_appium_selector_completions(needle, driver)
    except WebDriverException as e:
        return ["Exception (press esc to clear):", str(e)]

    # Return
    return [r[0] for r in results]


def get_autoit_selector_completions(needle, driver=AutoIt):
    if driver:
        results = _get_autoit_selector_completions(needle, driver)
        return [r for r in results if r]
    else:
        return []


def _get_white_selector_completions(needle, driver):
    if needle in ["ae=", "ae:"]:
        result = driver().pick()
        if result:
            return [
                completion
                for completion in [
                    f'id:{result["AutomationIdProperty"]}',
                    f'text:{result["NameProperty"]}',
                    f'help_text:{result["HelpTextProperty"]}',
                    f'class_name:{result["ClassNameProperty"]}',
                    f'control_type:{result["ControlTypeProperty"]}',
                ]
                if completion.split(":", 1)[-1]
            ]
        else:
            return []
    elif needle in ["template=", "template:"]:
        result = driver().pick(snip=True)
        if result and "bytes" in result:
            filename = f"{int(time.time())}.png"
            with open(os.path.join(os.getcwd(), filename), "bw") as fp:
                fp.write(result["bytes"])
            return [f"template=${{EXECDIR}}{os.path.sep * 2}{filename}"]
        else:
            return []

    else:
        return []


def get_white_selector_completions(needle, driver=PickSnipTool):
    # while only picker is implemented, we don't really use the needle
    if driver:
        results = _get_white_selector_completions(needle, driver)
        return [r for r in results if r]
    else:
        return []


def get_win32_selector_completions(needle):
    if needle.startswith("template"):
        filename = f"{int(time.time())}.bmp"
        region = ImageRegionTool(filename=os.path.join(os.getcwd(), filename)).run()
        if region is None:
            return []

        return [f"template=${{EXECDIR}}{os.path.sep * 2}{filename}"]

    elif needle.startswith("uia"):
        point = PointPickTool().run()
        if point is None:
            return []

        element = pywinauto.Desktop(backend="uia").from_point(*point)
        results = []
        for attribute in (
            "name",
            "rich_text",
            "automation_id",
            "control_type",
            "class_name",
        ):
            value = getattr(element, attribute, None)
            if callable(value):
                value = value()
            if value:
                results.append(f"{attribute}:{value}")

        return results

    else:
        return []


def _get_selenium_selector_completions(needle, driver):
    if IS_ID_SELECTOR_NEEDLE.match(needle):
        return get_selenium_id_selector_completions(needle, driver)
    elif IS_NAME_SELECTOR_NEEDLE.match(needle):
        return get_selenium_name_selector_completions(needle, driver)
    elif IS_CSS_SELECTOR_NEEDLE.match(needle):
        return get_selenium_css_selector_completions(needle, driver)
    elif IS_TAG_SELECTOR_NEEDLE.match(needle):
        return get_selenium_tag_selector_completions(needle, driver)
    elif IS_LINK_SELECTOR_NEEDLE.match(needle):
        return get_selenium_link_selector_completions(needle, driver)
    elif IS_XPATH_SELECTOR_NEEDLE.match(needle):
        return get_selenium_xpath_selector_completions(needle, driver)
    else:
        return []


def _get_appium_selector_completions(needle, driver):
    if IS_ID_SELECTOR_NEEDLE.match(needle):
        return get_appium_id_selector_completions(needle, driver)
    elif IS_XPATH_SELECTOR_NEEDLE.match(needle):
        return get_appium_xpath_selector_completions(needle, driver)
    else:
        return []


def _get_autoit_selector_completions(needle, driver):
    if needle.startswith("strTitle="):
        return get_autoit_win_selector_completions(needle, driver)
    elif needle.startswith("strControl="):
        return get_autoit_control_selector_completions(needle, driver)
    else:
        return []


def get_simmer_matches(elements, driver):
    matches = []
    if elements:
        script = "return ["
        script += ", ".join(
            [f"window.Simmer(arguments[{idx:d}])" for idx in range(len(elements))]
        )
        script += "];"
        results = driver.execute_script(script, *elements)
        for idx in range(len(results)):
            if results[idx]:
                matches.append((f"css:{results[idx]}", elements[idx]))
    return matches


def visible_or_all(results):
    return list(filter(lambda e: e.is_displayed(), results)) or results


def get_selenium_id_selector_completions(needle, driver):
    needle = needle[3:]
    matches = []
    if needle:
        results = driver.find_elements_by_css_selector(
            f'[id="{needle}"]'
        ) or driver.find_elements_by_css_selector(f'[id*="{needle}"]')
    else:
        results = driver.find_elements_by_xpath("//*[@id]")
    for result in visible_or_all(results):
        id_ = result.get_attribute("id")
        matches.append((f"id:{id_}", result))
    return matches


def get_appium_id_selector_completions(needle, driver):
    needle = needle[3:]
    matches = []
    if needle:
        results = driver.find_elements_by_xpath(
            f"//*[" f'contains(@resource-id, "{needle}")' f"]"
        )
    else:
        results = driver.find_elements_by_xpath("//*[@resource-id]")
    for result in visible_or_all(results):
        id_ = result.get_attribute("resource-id")
        matches.append((f'xpath=//*[@resource-id="{id_}"]', result))
    return matches


def get_selenium_name_selector_completions(needle, driver):
    needle = needle[5:]
    matches = []
    if needle:
        results = driver.find_elements_by_css_selector(
            f'[name="{needle}"]'
        ) or driver.find_elements_by_css_selector(f'[name*="{needle}"]')
    else:
        results = driver.find_elements_by_xpath("//*[@name]")
    for result in visible_or_all(results):
        name = result.get_attribute("name")
        matches.append((f"name:{name}", result))
    return matches


def get_selenium_needle_from_user(driver):
    try:
        return (
            driver.execute_async_script(
                """\
var node = document.getElementById('robotframework-interpreter-picker') ||
       document.createElement('div');
node.callback = arguments[arguments.length - 1];
node.setAttribute('id', 'robotframework-interpreter-picker');
node.setAttribute('onClick',
'this.parentNode.removeChild(this);' +
'this.callback(' +
'Simmer(' +
  'document.elementFromPoint(event.clientX, event.clientY)' +
')' +
');'
);
document.body.appendChild(node);
    """
            ) or
            ""
        )
    except TimeoutException:
        driver.execute_script(
            """\
var node = document.getElementById('robotframework-interpreter-picker');
if (node) { node.parentNode.removeChild(node); }
"""
        )
        return ""


def get_selenium_css_selector_completions(needle, driver):
    needle = needle[4:]
    unresolved = []
    results = []
    matches = []
    if not needle:
        needle = get_selenium_needle_from_user(driver)
    if needle:
        results = driver.find_elements_by_css_selector(needle)
    for result in visible_or_all(results):
        id_ = result.get_attribute("id")
        if " " in needle:  # always include simmer result for complex needles
            unresolved.append(result)
        if id_:
            matches.append((f"id:{id_}", result))
            continue
        if result.tag_name in FORM_TAG_NAMES and result.get_attribute("name"):
            name = result.get_attribute("name")
            matches.append((f"name:{name}", result))
            continue
        if result.tag_name == "a" and result.text:
            matches.append((f"link:{result.text}", result))
            continue
        if " " not in needle:
            unresolved.append(result)
    matches.extend(get_simmer_matches(unresolved, driver))
    return matches


def get_selenium_tag_selector_completions(needle, driver):
    needle = needle[4:]
    unresolved = []
    results = []
    matches = []
    if needle:
        results = driver.find_elements_by_css_selector(needle)
    for result in visible_or_all(results):
        id_ = result.get_attribute("id")
        if id_:
            matches.append((f"id:{id_}", result))
            continue
        if result.tag_name in FORM_TAG_NAMES:
            name = result.get_attribute("name")
            if name:
                matches.append((f"name:{name}", result))
                continue
        unresolved.append(result)
    matches.extend(get_simmer_matches(unresolved, driver))
    return matches


def get_selenium_link_selector_completions(needle, driver):
    needle = needle[5:]
    matches = []
    if needle:
        results = driver.find_elements_by_partial_link_text(needle)
    else:
        results = driver.find_elements_by_xpath("//a")
    for result in visible_or_all(results):
        if result.text:
            matches.append((f"link:{result.text}", result))
    return matches


def get_selenium_xpath_selector_completions(needle, driver):
    needle = needle[6:]
    results = []
    matches = []
    unresolved = []
    if needle:
        results = driver.find_elements_by_xpath(needle)
    for result in visible_or_all(results):
        id_ = result.get_attribute("id")
        if id_:
            matches.append((f"id:{id_}", result))
            continue
        if result.tag_name in FORM_TAG_NAMES:
            name = result.get_attribute("name")
            if name:
                matches.append((f"name:{name}", result))
                continue
        unresolved.append(result)
    matches.extend(get_simmer_matches(unresolved, driver))
    return matches


def get_appium_xpath_selector_completions(needle, driver):
    needle = needle[6:]
    matches = []
    if IS_TEXT.match(needle):
        results = driver.find_elements_by_xpath(
            f"//*[" f'contains(@text, "{needle}")' f"]"
        )
    elif needle:
        results = driver.find_elements_by_xpath(needle)
    else:
        results = driver.find_elements_by_xpath("//*[@resource-id]")
    for result in visible_or_all(results):
        id_ = result.get_attribute("resource-id")
        if id_:
            matches.append((f'xpath=//*[@resource-id="{id_}"]', result))
            continue
        class_ = result.get_attribute("class")
        text = result.get_attribute("text")
        matches.append((f'xpath=//{class_}[@text="{text}"]', result))
        continue

    return matches


def get_autoit_win_selector_completions(needle, driver):
    needle = needle[len("strTitle="):].strip()
    matches = []
    if needle:
        results = driver.WinList(needle)[0][1:]
        for result in results:
            matches.append(f"strTitle={result}  strText=")
    else:
        title = driver.WinGetTitle("")
        driver.ToolTip("Select window")
        while True:
            time.sleep(1)
            title_ = driver.WinGetTitle("")
            if title_ != title:
                driver.ToolTip("")
                matches.append(f"strTitle={title_}  strText=")
                break
    return matches


def get_autoit_control_selector_completions(needle, driver):  # noqa: C901
    needle = needle[len("strControl="):].strip()
    if needle:
        return []

    matches = []
    title = driver.WinGetTitle("")
    driver.ToolTip("Select control")

    while True:
        time.sleep(1)
        title_ = driver.WinGetTitle("")
        if title_ != title:
            driver.ToolTip("")

            pointer_x = driver.MouseGetPosX()
            pointer_y = driver.MouseGetPosY()
            counter = {}
            match = None

            time.sleep(1)

            for class_ in driver.WinGetClassList(title_).split():
                counter.setdefault(class_, 0)
                counter[class_] += 1
                ref = f"{class_}{counter[class_]:d}"
                handle = driver.ControlGetHandle(title_, "", ref)

                x = driver.WinGetPosX(f"[HANDLE:{handle}]")
                if x > pointer_x:  # control begins too late x
                    continue

                w = driver.ControlGetPosWidth(title_, "", ref)
                if (x + w) < pointer_x:  # control ends too early x
                    continue

                y = driver.WinGetPosY(f"[HANDLE:{handle}]")
                if y > pointer_y:  # control begins too late y
                    continue

                h = driver.ControlGetPosHeight(title_, "", ref)
                if (y + h) < pointer_y:  # control ends too late y
                    continue

                if match and w * h > match[1]:
                    continue

                match = [ref, w * h]

            driver.ToolTip("")

            if match:
                matches.append(f"strControl={match[0]}")

            break
    return matches



import inspect

from robot.errors import DataError
from robot.libdocpkg import LibraryDocumentation
from robot.libraries.BuiltIn import BuiltIn

from .utils import lunr_builder, to_mime_and_metadata
from .constants import CONTEXT_LIBRARIES


class RobotKeywordsIndexerListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self):
        self.builder = lunr_builder("dottedname", ["dottedname", "name"])
        self.index = None
        self.libraries = []
        self.keywords = {}

        self.library_import("BuiltIn", {})
        for name, keywords in CONTEXT_LIBRARIES.items():
            self._library_import(keywords, name)

    def library_import(self, alias, attributes):
        name = attributes.get("originalName") or alias

        if alias not in self.libraries:
            self.libraries.append(alias)
            try:
                lib_doc = LibraryDocumentation(name)
                self._library_import(lib_doc, alias)
            except DataError:
                pass

    def _library_import(self, lib_doc, alias):
        if isinstance(lib_doc, list):
            keywords = lib_doc
            doc_format = "REST"
        else:
            keywords = lib_doc.keywords
            doc_format = lib_doc.doc_format
        for keyword in keywords:
            keyword.doc_format = doc_format
            self.builder.add(
                {"name": keyword.name, "dottedname": f"{alias}.{keyword.name}"}
            )
            self.keywords[f"{alias}.{keyword.name}"] = keyword
        if len(self.keywords):
            self.index = self.builder.build()

    def resource_import(self, name, attributes):
        if name not in self.libraries:
            self.libraries.append(name)
            try:
                resource_doc = LibraryDocumentation(name)
                self._resource_import(resource_doc.keywords)
            except DataError:
                pass

    def _resource_import(self, keywords):
        for keyword in keywords:
            keyword.doc_format = "REST"
            self.builder.add(
                {"name": keyword.name, "dottedname": keyword.name}
            )
            self.keywords[keyword.name] = keyword
        if len(self.keywords):
            self.index = self.builder.build()

    def import_from_suite_data(self, suite):
        self._resource_import(suite.resource.keywords)
        try:
            for import_data in suite.resource.imports:
                attributes = {}
                if import_data.type == "Library":
                    alias = import_data.alias or import_data.name
                    attributes["originalName"] = import_data.name
                    self.library_import(alias, attributes)
                else:
                    name = import_data.name
                    self.resource_import(name, attributes)
        except AttributeError:
            pass


class StatusEventListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, callback=None):
        self.callback = callback

    def start_test(self, name, attributes):
        if self.callback is not None:
            self.callback({"test": name})

    def start_keyword(self, name, attributes):
        if self.callback is not None:
            self.callback({"keyword": name})


class ReturnValueListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self):
        self.return_value = None

    def end_keyword(self, name, attributes):
        frame = inspect.currentframe()
        while frame is not None:
            if "return_value" in frame.f_locals:
                self.return_value = frame.f_locals.get("return_value")
                break
            frame = frame.f_back

    def start_test(self, name, attributes):
        self.return_value = None

    def get_last_value(self):
        if self.return_value is not None and self.return_value != "" and self.return_value != b"":
            value = to_mime_and_metadata(self.return_value)

            self.return_value = None
            return value
        return None


def clear_drivers(drivers, type_):
    remained = []
    for driver in drivers:
        if driver.get("type") != type_:
            remained.append(driver)
    drivers.clear()
    drivers.extend(remained)


def get_webdrivers(cache, type_):
    drivers = []
    for idx in range(len(cache._connections)):
        conn = cache._connections[idx]
        if conn in cache._closed:
            continue
        aliases = []
        for alias, idx_ in cache._aliases.items():
            if (idx + 1) == idx_:
                aliases.append(alias)
        drivers.append(
            dict(
                instance=conn,
                aliases=aliases,
                current=conn == cache.current,
                type=type_,
            )
        )
    return drivers


def set_webdrivers(drivers, cache, type_):
    idx = 1
    for driver in drivers:
        if driver["type"] != type_:
            continue
        cache._connections.append(driver["instance"])
        for alias in driver["aliases"]:
            cache._aliases[alias] = idx
        if driver["current"]:
            cache.current = driver["instance"]
        idx += 1


class SeleniumConnectionsListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, drivers: list):
        self.drivers = drivers

    def end_suite(self, name, attributes):
        try:
            builtin = BuiltIn()
            try:
                instance = builtin.get_library_instance("SeleniumLibrary")
            except RuntimeError:
                instance = builtin.get_library_instance("Selenium2Library")
            clear_drivers(self.drivers, "selenium")
            self.drivers.extend(get_webdrivers(instance._drivers, "selenium"))
        except RuntimeError:
            pass

    def start_suite(self, name, attributes):
        try:
            builtin = BuiltIn()
            try:
                instance = builtin.get_library_instance("SeleniumLibrary")
            except RuntimeError:
                instance = builtin.get_library_instance("Selenium2Library")
            set_webdrivers(self.drivers, instance._drivers, "selenium")
        except RuntimeError:
            pass


class RpaBrowserConnectionsListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, drivers: list):
        self.drivers = drivers

    def end_suite(self, name, attributes):
        try:
            instance = BuiltIn().get_library_instance("RPA.Browser")
            clear_drivers(self.drivers, "RPA.Browser")
            self.drivers.extend(get_webdrivers(instance._drivers, "RPA.Browser"))
        except RuntimeError:
            pass

    def start_suite(self, name, attributes):
        try:
            instance = BuiltIn().get_library_instance("RPA.Browser")
            set_webdrivers(self.drivers, instance._drivers, "RPA.Browser")
        except RuntimeError:
            pass


class JupyterConnectionsListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, drivers: list):
        self.drivers = drivers

    def end_suite(self, name, attributes):
        try:
            builtin = BuiltIn()
            instance = builtin.get_library_instance("JupyterLibrary")
            clear_drivers(self.drivers, "jupyter")
            self.drivers.extend(get_webdrivers(instance._drivers, "jupyter"))
        except RuntimeError:
            pass

    def start_suite(self, name, attributes):
        try:
            builtin = BuiltIn()
            instance = builtin.get_library_instance("JupyterLibrary")
            set_webdrivers(self.drivers, instance._drivers, "jupyter")
        except RuntimeError:
            pass


class AppiumConnectionsListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, drivers: list):
        self.drivers = drivers

    def end_suite(self, name, attributes):
        try:
            builtin = BuiltIn()
            instance = builtin.get_library_instance("AppiumLibrary")
            clear_drivers(self.drivers, "appium")
            self.drivers.extend(get_webdrivers(instance._cache, "appium"))
        except RuntimeError:
            pass

    def start_suite(self, name, attributes):
        try:
            builtin = BuiltIn()
            instance = builtin.get_library_instance("AppiumLibrary")
            set_webdrivers(self.drivers, instance._cache, "appium")
        except RuntimeError:
            pass


class WhiteLibraryListener:
    ROBOT_LISTENER_API_VERSION = 2

    def __init__(self, drivers: list):
        self.drivers = drivers

    def end_suite(self, name, attributes):
        try:
            builtin = BuiltIn()
            instance = builtin.get_library_instance("WhiteLibrary")
            clear_drivers(self.drivers, "white")
            self.drivers.append(
                dict(
                    instance=(
                        getattr(instance, "app", None),
                        getattr(instance, "window", None),
                        getattr(instance, "screenshotter", None),
                    ),
                    aliases=[],
                    current=True,
                    type="white",
                )
            )
        except RuntimeError:
            pass

    def start_suite(self, name, attributes):
        try:
            builtin = BuiltIn()
            instance = builtin.get_library_instance("WhiteLibrary")
            for driver in self.drivers:
                if driver.get("type") == "white" and driver.get("current"):
                    setattr(instance, "app", driver["instance"][0])
                    setattr(instance, "window", driver["instance"][1])
                    setattr(instance, "screenshotter", driver["instance"][2])
        except RuntimeError:
            pass



from .listeners import (   # noqa
    RobotKeywordsIndexerListener, SeleniumConnectionsListener, RpaBrowserConnectionsListener,
    JupyterConnectionsListener, AppiumConnectionsListener, WhiteLibraryListener,
    ReturnValueListener, StatusEventListener
)
from .interpreter import init_suite, execute, complete, inspect, shutdown_drivers, ProgressUpdater  # noqa



import os
from io import BytesIO
import base64
import binascii
import urllib
from urllib.parse import unquote
import mimetypes
import re
import json
from json import JSONDecodeError
from typing import List
from difflib import SequenceMatcher
from copy import deepcopy
from operator import itemgetter

from robot.libraries import STDLIBS
from robot.libdocpkg.htmlwriter import DocToHtml

from PIL import Image

from lunr.builder import Builder
from lunr.stemmer import stemmer
from lunr.stop_word_filter import stop_word_filter
from lunr.trimmer import trimmer

from pygments.formatters import HtmlFormatter
from pygments.lexers import get_lexer_by_name
import pygments

from .constants import SCRIPT_DISPLAY_LOG, NAME_REGEXP


def data_uri(mimetype, data):
    return "data:{};base64,{}".format(mimetype, base64.b64encode(data).decode("utf-8"))


def process_screenshots(outputdir: str):
    cwd = os.getcwd()

    with open(os.path.join(outputdir, "output.xml")) as fp:
        xml = fp.read()

    for src in re.findall('img src="([^"]+)', xml):
        if os.path.exists(src):
            filename = src
        elif os.path.exists(os.path.join(outputdir, src)):
            filename = os.path.join(outputdir, src)
        elif os.path.exists(os.path.join(cwd, src)):
            filename = os.path.join(cwd, src)
        elif src.startswith("data:"):
            filename = None
            try:
                spec, uri = src.split(",", 1)
                spec, encoding = spec.split(";", 1)
                spec, mimetype = spec.split(":", 1)
                if not (encoding == "base64" and mimetype.startswith("image/")):
                    continue
                data = base64.b64decode(unquote(uri).encode("utf-8"))
                im = Image.open(BytesIO(data))
            except (binascii.Error, IndexError, ValueError):
                continue
        else:
            continue
        if filename:
            im = Image.open(filename)
            mimetype = Image.MIME[im.format]
            # Fix issue where Pillow on Windows returns APNG for PNG
            if mimetype == "image/apng":
                mimetype = "image/png"
            with open(filename, "rb") as fp:
                data = fp.read()
        uri = data_uri(mimetype, data)
        xml = xml.replace('a href="{}"'.format(src), "a")
        xml = xml.replace(
            'img src="{}" width="800px"'.format(src),
            'img src="{}" style="max-width:800px;"'.format(uri),
        )
        xml = xml.replace('img src="{}"'.format(src), 'img src="{}"'.format(uri))

    with open(os.path.join(outputdir, "output.xml"), "w") as fp:
        fp.write(xml)


def display_log(html, filename=""):
    if isinstance(html, str):
        html = html.encode("utf-8")
    return SCRIPT_DISPLAY_LOG.format(
        content=base64.b64encode(html).decode("utf-8"), filename=filename
    )


def highlight(language, data):
    lexer = get_lexer_by_name(language)
    formatter = HtmlFormatter(noclasses=True, nowrap=True)
    return pygments.highlight(data, lexer, formatter)


def to_html(obj):
    """Return object as highlighted JSON."""
    return highlight("json", json.dumps(obj, sort_keys=False, indent=4))


def img_to_data(path):
    mime, _ = mimetypes.guess_type(path)

    if path.startswith("http"):
        with urllib.request.urlopen(path) as fp:
            data = fp.read()
    else:
        with open(path, 'rb') as fp:
            data = fp.read()

    return data_uri(mime, data)


def to_mime_and_metadata(obj):
    if isinstance(obj, bytes):
        obj = base64.b64encode(obj).decode("utf-8")
        return {"text/html": to_html(obj)}
    elif isinstance(obj, str) and (obj.startswith("http") or os.path.exists(obj)):
        if re.match(r".*\.(gif|jpg|svg|jpeg||png)$", obj, re.I):
            return {"text/html": "<img src='{}'>".format(img_to_data(obj))}
    elif hasattr(obj, "_repr_mimebundle_"):
        obj.embed = True
        return obj._repr_mimebundle_()
    elif hasattr(obj, "_repr_json_"):
        obj.embed = True
        return {"application/json": obj._repr_json_()}
    elif hasattr(obj, "_repr_html_"):
        obj.embed = True
        return {"text/html": obj._repr_html_()}
    elif hasattr(obj, "_repr_png_"):
        return {"image/png": obj._repr_png_()}
    elif hasattr(obj, "_repr_jpeg_"):
        return {"image/jpeg": obj._repr_jpeg_()}
    elif hasattr(obj, "_repr_svg_"):
        return {"image/svg": obj._repr_svg_()}
    try:
        if isinstance(obj, str):
            return {"text/html": f"<pre>{to_html(obj)}</pre>".replace("\\n", "\n")}
    except (TypeError, JSONDecodeError):
        pass
    try:
        return {"text/html": to_html(obj)}
    except TypeError:
        return {}


def detect_robot_context(code: str, cursor_pos: int):
    """Return robot code context in cursor position."""
    code = code[:cursor_pos]
    line = code.rsplit("\n")[-1]
    context_parts = code.rsplit("***", 2)
    if len(context_parts) != 3:
        return "__root__"
    else:
        context_name = context_parts[1].strip().lower()
        if context_name == "settings":
            return "__settings__"
        elif line.lstrip() == line:
            return "__root__"
        elif context_name in ["tasks", "test cases"]:
            return "__tasks__"
        elif context_name == "keywords":
            return "__keywords__"
        else:
            return "__root__"


def line_at_cursor(code: str, cursor_pos: int = 0):
    """Return the line of code that is at the cursor position."""
    offset = 0
    lines = code.splitlines(True)
    for line in lines:
        next_offset = offset + len(line)
        if not line.endswith('\n'):
            # If the last line doesn't have a trailing newline, treat it as if
            # it does so that the cursor at the end of the line still counts
            # as being on that line.
            next_offset += 1
        if next_offset > cursor_pos:
            break
        offset = next_offset
    else:
        line = ""
    return (line, offset)


def scored_results(needle: str, results: List[str]) -> List[str]:
    results = deepcopy(results)
    for result in results:
        match = SequenceMatcher(
            None, needle.lower(), result["ref"].lower(), autojunk=False
        ).find_longest_match(0, len(needle), 0, len(result["ref"]))
        result["score"] = (match.size, match.size / float(len(result["ref"])))
    return list(reversed(sorted(results, key=itemgetter("score"))))


def remove_prefix(value, prefix):
    if value.startswith(prefix):
        value = value[len(prefix):]
    return value


def complete_libraries(needle: str, extra_libraries: List[str]) -> List[str]:
    """Complete library names."""
    matches = []

    libs = list(STDLIBS) + extra_libraries

    for lib in libs:
        if lib.lower().startswith(needle):
            matches.append(lib)

    return matches


def readable_keyword(s):
    """Return keyword in title case."""
    if s and not s.startswith("*") and not s.startswith("["):
        if s.count("."):
            library, name = s.rsplit(".", 1)
            return library + "." + name[0:].title()
        else:
            return s
    else:
        return s


def lunr_builder(ref, fields):
    """A convenience function to configure and construct a lunr.Builder.

    Returns:
        Index: The populated Index ready to search against.
    """
    builder = Builder()
    builder.pipeline.add(trimmer, stop_word_filter, stemmer)
    builder.search_pipeline.add(stemmer)
    builder.ref(ref)
    for field in fields:
        builder.field(field)
    return builder


def lunr_query(query):
    query = re.sub(r"([:*])", r"\\\1", query, re.U)
    query = re.sub(r"[\[\]]", r"", query, re.U)
    return f"*{query.strip().lower()}*"


def get_lunr_completions(needle: str, index, keywords, context):
    matches = []
    results = []

    if needle.rstrip():
        query = lunr_query(needle)
        results = index.search(query)
        results += index.search(query.strip("*"))

    for result in scored_results(needle, results):
        ref = result["ref"]
        if ref.startswith("__") and not ref.startswith(context):
            continue
        if not ref.startswith(context) and context not in [
            "__tasks__",
            "__keywords__",
            "__settings__",
        ]:
            continue
        if not needle.count("."):
            keyword = keywords[ref].name
            if keyword not in matches:
                matches.append(readable_keyword(keyword))
        else:
            matches.append(readable_keyword(ref))
    return matches


def get_keyword_doc(keyword):
    title = keyword.name.strip("*").strip()
    title_html = f"<strong>{title}</strong>"
    if keyword.args:
        title += " " + ", ".join(keyword.args)
        title_html += " " + ", ".join(keyword.args)

    body = ""
    if keyword.doc:
        body = "\n\n" + keyword.doc

    return {
        "text/plain": title + "\n\n" + body,
        "text/html": f"<p>{title_html}</p>" +
        NAME_REGEXP.sub(
            lambda m: f"<code>{m.group(1)}</code>", DocToHtml(keyword.doc_format)(body)
        ),
    }



#!/usr/bin/env python

from setuptools import setup, find_packages

__AUTHOR__ = 'QuantStack dev team'

setup(
    name='robotframework-interpreter',
    version='0.6.1',
    description='Utility functions for building a Robot Framework interpreter.',
    author=__AUTHOR__,
    maintainer=__AUTHOR__,
    url='https://github.com/martinRenou/robotframework-interpreter',
    license='BSD 3-Clause',
    keywords='robotframework interpreter',
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'robotframework>=3.2,<4',
        'lunr',
        'pillow',
        'pygments',
        'ipywidgets'
    ],
    extras_require={
        'testing': ['flake8'],
    },
    platforms=['any'],
    classifiers=[
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import logging
import time
from abc import ABC, abstractmethod
from ctypes import windll

try:
    windll.user32.SetProcessDPIAware()
except AttributeError:
    pass  # Might not be supported

import win32api
import win32con
import win32event
import win32gui
import win32ui


def lparam_to_point(lparam):
    """Parse coordinates from window message."""
    x = win32api.LOWORD(lparam)
    y = win32api.HIWORD(lparam)

    # Coordinates are signed, and can be negative
    # if on a second monitor
    if x & 32768:
        x |= -65536
    if y & 32768:
        y |= -65536

    return x, y


class SelectorToolBase(ABC):
    """Base class for selector tools.

    Displays a full-screen screenshot of the current desktop, and allows
    drawing on top of it and capturing the mouse position/clicks.
    """

    CURSOR = win32con.IDC_CROSS
    STOP_EVENT = win32event.CreateEvent(None, 0, 0, None)

    def __init__(self, timeout=30):
        self.logger = logging.getLogger(__name__)
        self.timeout = float(timeout)
        self._screenshot = None
        self._background = None
        self._window = None

    @property
    @abstractmethod
    def message_map(self):
        """Map of window message types to handler functions."""
        raise NotImplementedError

    @property
    @abstractmethod
    def result(self):
        """Result of the operation. Should be set before ``stop()`` is called."""
        raise NotImplementedError

    def run(self):
        """Create window and start handling messages.
        Blocks until ``stop()``` is called or timeout is reached.
        """
        self._create_window()
        try:
            timeout = time.time() + self.timeout
            while time.time() < timeout:
                if not self._pump_messages():
                    break
            else:
                self.logger.warning("Timeout reached")
        finally:
            self._remove_window()

        return self.result

    def stop(self):
        """Signal message loop to stop processing."""
        win32event.SetEvent(self.STOP_EVENT)

    def _pump_messages(self):
        """Handle messages from the Windows message loop.
        Returns `False` when process should stop.

        Notes:
            - WAIT_OBJECT_0 refers to the first item in given waitables
            - WAIT_OBJECT_0 + len(waitables) refers to a message defined by
              the mask, i.e. QS_ALLEVENTS
        """
        waitables = [self.STOP_EVENT]
        event = win32event.MsgWaitForMultipleObjects(
            waitables, False, 100, win32event.QS_ALLEVENTS,
        )

        if event == win32event.WAIT_OBJECT_0:
            return False
        if event == win32event.WAIT_OBJECT_0 + len(waitables):
            win32gui.PumpWaitingMessages()
            return True
        if event == win32event.WAIT_TIMEOUT:
            return True

        raise RuntimeError(f"Unexpected event: {event}")

    def _create_window(self):
        """Create full-screen window with screenshot as background."""
        # NB: For some reason a reference to the brush needs to be kept in Python,
        #     or otherwise there will only be a white background
        self._screenshot = self._take_screenshot()
        self._background = win32gui.CreatePatternBrush(self._screenshot.GetHandle())

        instance = win32api.GetModuleHandle()
        wndclass = win32gui.WNDCLASS()

        # Combine default message handlers with child class
        message_map = {win32con.WM_DESTROY: self._on_destroy}
        message_map.update(self.message_map)

        wndclass.style = win32con.CS_HREDRAW | win32con.CS_VREDRAW
        wndclass.lpfnWndProc = message_map
        wndclass.hInstance = instance
        wndclass.hbrBackground = self._background
        wndclass.hCursor = win32gui.LoadCursor(None, self.CURSOR)
        wndclass.lpszClassName = self.__class__.__name__

        wndclass_atom = win32gui.RegisterClass(wndclass)

        # WS_POPUP:         The window is a pop-up window
        # WS_VISIBLE:       The window is initially visible
        style = win32con.WS_POPUP | win32con.WS_VISIBLE

        # WS_EX_COMPOSITED: Paints all descendants in bottom-to-top order
        # WS_EX_LAYERED:    Painting is buffered to bitmap (faster)
        # WS_EX_TOOLWINDOW: Tool window, i.e. does not appear in the taskbar
        # WS_EX_TOPMOST:    Stay on top of all non-topmost windows
        ex_style = (
            win32con.WS_EX_COMPOSITED |
            win32con.WS_EX_LAYERED |
            win32con.WS_EX_TOOLWINDOW |
            win32con.WS_EX_TOPMOST
        )

        window = win32gui.CreateWindowEx(
            ex_style,  # extended window style
            wndclass_atom,  # class name
            None,  # window title
            style,  # window style
            0,  # x
            0,  # y
            win32api.GetSystemMetrics(win32con.SM_CXSCREEN),  # width
            win32api.GetSystemMetrics(win32con.SM_CYSCREEN),  # height
            None,  # parent
            None,  # menu
            instance,  # instance
            None,  # reserved
        )

        win32gui.ShowWindow(window, win32con.SW_SHOW)
        self._window = window

    def _remove_window(self):
        """Destroy window and free reserved resources."""
        if self._window:
            win32gui.DestroyWindow(self._window)
            win32gui.UnregisterClass(self.__class__.__name__, None)
            self._window = None

        if self._screenshot:
            win32gui.DeleteObject(self._screenshot.GetHandle())
            self._screenshot = None

    def _on_destroy(self, hwnd, message, wparam, lparam):
        """Default handler for WM_DESTROY."""
        win32gui.PostQuitMessage(0)
        return 0

    @staticmethod
    def _take_screenshot():
        """Bit-block copy current desktop pixel data into a bitmap."""
        src = win32gui.GetDesktopWindow()
        src_hdc = win32gui.GetWindowDC(src)
        src_dc = win32ui.CreateDCFromHandle(src_hdc)

        src_width = win32api.GetSystemMetrics(win32con.SM_CXSCREEN)
        src_height = win32api.GetSystemMetrics(win32con.SM_CYSCREEN)

        dst_bitmap = win32ui.CreateBitmap()
        dst_bitmap.CreateCompatibleBitmap(src_dc, src_width, src_height)

        dst_dc = src_dc.CreateCompatibleDC()
        dst_default = dst_dc.SelectObject(dst_bitmap)
        assert dst_default is not None

        dst_dc.BitBlt((0, 0), (src_width, src_height), src_dc, (0, 0), win32con.SRCCOPY)

        # Default bitmap should be restored to DC after use,
        # otherwise some memory/performance issues might occur
        dst_dc.SelectObject(dst_default)

        dst_dc.DeleteDC()
        src_dc.DeleteDC()
        win32gui.ReleaseDC(src, src_hdc)

        return dst_bitmap


class ImageRegionTool(SelectorToolBase):
    """Draw a selection rectangle and return the coordinates.
    Optionally save the region as an image file.
    """

    REGION_WIDTH = 3
    REGION_COLOR = win32api.RGB(255, 0, 0)

    def __init__(self, filename=None, timeout=30):
        super().__init__(timeout)
        self.filename = filename
        self._selection_start = None
        self._selection = None

    @property
    def message_map(self):
        return {
            win32con.WM_LBUTTONDOWN: self.on_lbuttondown,
            win32con.WM_MOUSEMOVE: self.on_mousemove,
            win32con.WM_LBUTTONUP: self.on_lbuttonup,
            win32con.WM_KEYDOWN: self.on_keydown,
            win32con.WM_PAINT: self.on_paint,
        }

    @property
    def result(self):
        return self._selection

    def _save_selection(self):
        """Save selection region from the original screenshot."""
        assert self.filename is not None
        assert self._selection is not None

        left, top, right, bottom = self._selection
        width = right - left
        height = bottom - top

        src_dch = win32gui.CreateCompatibleDC(None)
        src_dc = win32ui.CreateDCFromHandle(src_dch)

        src_default = src_dc.SelectObject(self._screenshot)
        assert src_default is not None

        dst_dch = win32gui.CreateCompatibleDC(None)
        dst_dc = win32ui.CreateDCFromHandle(dst_dch)

        dst_bitmap = win32ui.CreateBitmap()
        dst_bitmap.CreateCompatibleBitmap(src_dc, width, height)

        dst_default = dst_dc.SelectObject(dst_bitmap)
        assert dst_default is not None

        dst_dc.BitBlt((0, 0), (width, height), src_dc, (left, top), win32con.SRCCOPY)

        dst_bitmap.SaveBitmapFile(dst_dc, self.filename)
        self.logger.info("Saved selected region as '%s'", self.filename)

        dst_dc.SelectObject(dst_default)
        src_dc.SelectObject(src_default)

        win32gui.DeleteObject(dst_bitmap.GetHandle())
        dst_dc.DeleteDC()
        src_dc.DeleteDC()

    def on_lbuttondown(self, hwnd, message, wparam, lparam):
        """Left button being pressed starts new selection."""
        self._selection_start = lparam_to_point(lparam)
        self._selection = None

        # Force re-draw
        win32gui.InvalidateRect(hwnd, None, True)
        return 0

    def on_mousemove(self, hwnd, message, wparam, lparam):
        """Update selection region with mouse location."""
        # Only update region while left button is pressed
        if not wparam & win32con.MK_LBUTTON:
            return 0

        x1, y1 = self._selection_start
        x2, y2 = lparam_to_point(lparam)

        # Handle drawing region from bottom-right to top-left
        self._selection = [min(x1, x2), min(y1, y2), max(x1, x2), max(y1, y2)]

        # Force re-draw
        win32gui.InvalidateRect(hwnd, None, True)
        return 0

    def on_lbuttonup(self, hwnd, message, wparam, lparam):
        """Selection is finished when left button is no longer pressed."""
        if self.filename and self._selection:
            self._save_selection()

        self.stop()
        return 0

    def on_keydown(self, hwnd, message, wparam, lparam):
        """Cancel selection with escape key."""
        if wparam == win32con.VK_ESCAPE:
            self._selection = None
            self.stop()
        return 0

    def on_paint(self, hwnd, message, wparam, lparam):
        """Draw current selection rectangle."""
        hdc, paint = win32gui.BeginPaint(hwnd)

        # Selection not started yet
        if self._selection is None:
            win32gui.EndPaint(hwnd, paint)
            return 0

        brush = win32gui.GetStockObject(win32con.NULL_BRUSH)
        pen = win32gui.CreatePen(
            win32con.PS_SOLID, self.REGION_WIDTH, self.REGION_COLOR
        )

        brush_default = win32gui.SelectObject(hdc, brush)
        pen_default = win32gui.SelectObject(hdc, pen)

        win32gui.Rectangle(hdc, *self._selection)

        win32gui.SelectObject(hdc, pen_default)
        win32gui.SelectObject(hdc, brush_default)

        win32gui.EndPaint(hwnd, paint)
        return 0


class PointPickTool(SelectorToolBase):
    """Pick a point on the screen and return the coordinates."""

    def __init__(self, timeout=30):
        super().__init__(timeout)
        self._point = None

    @property
    def message_map(self):
        return {win32con.WM_LBUTTONDOWN: self.on_lbuttondown}

    @property
    def result(self):
        return self._point

    def on_lbuttondown(self, hwnd, message, wparam, lparam):
        """Store clicked point."""
        self._point = lparam_to_point(lparam)
        self.stop()
        return 0



from System.Drawing import Bitmap
from System.Drawing import Color
from System.Drawing import Graphics
from System.Drawing import GraphicsUnit
from System.Drawing import Pen
from System.Drawing import Rectangle
from System.Drawing import Size
from System.Drawing import SolidBrush
from System.Drawing.Imaging import ImageFormat
from System.Drawing.Imaging import PixelFormat
from System.IO import MemoryStream
from System.Windows import Point
from System.Windows.Automation import AutomationElement
from System.Windows.Forms import Cursors
from System.Windows.Forms import DialogResult
from System.Windows.Forms import Form
from System.Windows.Forms import FormBorderStyle
from System.Windows.Forms import FormWindowState
from System.Windows.Forms import Keys
from System.Windows.Forms import Screen
from System.Windows.Forms import Timer
from TestStack.White.InputDevices import Mouse
import os
import time


class PickSnipTool(Form):

    mouse_down = None
    mouse_down_button = None
    mouse_down_seconds = 0
    mouse_down_timer = None
    mouse_up = None

    snip_enabled = False
    snip_rectangle = None

    click_timeout = 1  # set this to eg. 4 to enable click + pick

    @staticmethod
    def take_screenshot(as_bytes=False):
        bounds = Screen.PrimaryScreen.Bounds
        screenshot = Bitmap(bounds.Width, bounds.Height, PixelFormat.Format32bppPArgb)
        graphics = Graphics.FromImage(screenshot)
        graphics.CopyFromScreen(0, 0, 0, 0, screenshot.Size)
        if as_bytes:
            fp = MemoryStream()
            screenshot.Save(fp, ImageFormat.Png)
            return bytes(bytearray(fp.ToArray()))
        else:
            return screenshot

    @staticmethod
    def pick(snip=False):
        snipper = PickSnipTool(PickSnipTool.take_screenshot(), snip=snip)

        while True:
            result = snipper.ShowDialog()
            if result in [DialogResult.OK, DialogResult.Cancel]:
                break
            if snipper.mouse_down_seconds == 1:
                Mouse.Instance.Click(snipper.mouse_down)
                time.sleep(0.5)
            snipper.BackgroundImage = PickSnipTool.take_screenshot()

        if result == DialogResult.OK and snip:
            if snipper.snip_rectangle.Width and snipper.snip_rectangle.Height:
                img = Bitmap(
                    snipper.snip_rectangle.Width, snipper.snip_rectangle.Height
                )
                gr = Graphics.FromImage(img)
                gr.DrawImage(
                    snipper.BackgroundImage,
                    Rectangle(0, 0, img.Width, img.Height),
                    snipper.snip_rectangle,
                    GraphicsUnit.Pixel,
                )
                fp = MemoryStream()
                img.Save(fp, ImageFormat.Png)
                return {"bytes": bytes(bytearray(fp.ToArray()))}
            return {}

        elif result == DialogResult.OK:
            if (
                snipper.mouse_down_seconds and
                snipper.mouse_down_seconds <= snipper.click_timeout
            ):
                Mouse.Instance.Click(snipper.mouse_down)
                time.sleep(0.5)
            el = AutomationElement.FromPoint(snipper.mouse_up)
            result = {
                prop.ProgrammaticName.split(".", 1)[-1]: el.GetCurrentPropertyValue(
                    prop
                )
                for prop in el.GetSupportedProperties()
            }
            result.update(
                {
                    "NameProperty": el.GetCurrentPropertyValue(el.NameProperty),
                    "ControlTypeProperty": el.GetCurrentPropertyValue(
                        el.ControlTypeProperty
                    ).ProgrammaticName.split(".", 1)[-1],
                    "AutomationIdProperty": el.GetCurrentPropertyValue(
                        el.AutomationIdProperty
                    ),
                }
            )
            return result
        else:
            return {}

    def __init__(self, screenshot=None, snip=False):
        super(PickSnipTool, self).__init__()
        self.snip_enabled = snip

        self.Cursor = Cursors.Cross
        self.BackgroundImage = screenshot
        self.ShowInTaskbar = False
        self.FormBorderStyle = getattr(FormBorderStyle, "None")
        self.WindowState = FormWindowState.Maximized
        self.DoubleBuffered = True
        self.TopMost = True
        self.TopLevel = True

        self.MouseDown += self.on_mouse_down
        self.MouseMove += self.on_mouse_move
        self.MouseUp += self.on_mouse_up
        self.KeyUp += self.on_key_up
        self.KeyDown += self.OnKeyDown
        self.Paint += self.on_paint

    def on_timer_tick(self, sender, e):
        self.mouse_down_seconds += 1
        if self.mouse_down_seconds == 1:
            self.DialogResult = DialogResult.Retry
        if self.mouse_down_seconds > self.click_timeout:
            self.DialogResult = DialogResult.Retry

    def on_mouse_down(self, sender, e):
        self.mouse_down = Point(e.Location.X, e.Location.Y)
        self.mouse_down_button = e.Button
        self.mouse_down_seconds = 0
        if not self.snip_enabled:
            self.mouse_down_timer = Timer()
            self.mouse_down_timer.Interval = 1000
            self.mouse_down_timer.Tick += self.on_timer_tick
            self.mouse_down_timer.Start()
        self.snip_rectangle = Rectangle(e.Location, Size(0, 0))

    def on_mouse_move(self, sender, e):
        if self.mouse_down and self.snip_enabled:
            x1 = int(min(e.X, self.mouse_down.X))
            y1 = int(min(e.Y, self.mouse_down.Y))
            x2 = int(max(e.X, self.mouse_down.X))
            y2 = int(max(e.Y, self.mouse_down.Y))
            self.snip_rectangle = Rectangle(x1, y1, x2 - x1, y2 - y1)
            self.Invalidate()

    def on_mouse_up(self, sender, e):
        self.mouse_up = Point(e.Location.X, e.Location.Y)
        if self.mouse_down_timer is not None:
            self.mouse_down_timer.Stop()
        self.DialogResult = DialogResult.OK

    def on_paint(self, sender, e):
        if self.snip_enabled and self.snip_rectangle is not None:
            br = SolidBrush(Color.FromArgb(120, Color.White))
            rc = self.snip_rectangle
            pen = Pen(Color.Red, 3.0)

            x1 = rc.X
            x2 = rc.X + rc.Width
            y1 = rc.Y
            y2 = rc.Y + rc.Height

            e.Graphics.FillRectangle(br, Rectangle(0, 0, x1, self.Height))
            e.Graphics.FillRectangle(br, Rectangle(x2, 0, self.Width - x2, self.Height))
            e.Graphics.FillRectangle(br, Rectangle(x1, 0, x2 - x1, y1))
            e.Graphics.FillRectangle(br, Rectangle(x1, y2, x2 - x1, self.Height - y2))
            e.Graphics.DrawRectangle(pen, rc)

    def on_key_up(self, sender, e):
        if e.KeyCode == Keys.Escape:
            self.DialogResult = DialogResult.Cancel


class WhiteLibraryCompanion:
    """Complementary keyword library to use OpenCV based image recognition with WhiteLibrary.
    """

    def match_template(self, template: str, similarity: float = 0.8):
        import numpy as np  # noqa
        import cv2 as cv  # noqa

        screenshot = PickSnipTool.take_screenshot(as_bytes=True)
        image = cv.imdecode(np.frombuffer(screenshot, np.uint8), cv.IMREAD_COLOR)
        template = cv.imread(template)
        result = cv.matchTemplate(image, template, cv.TM_CCOEFF_NORMED)
        min_val, max_val, min_loc, max_loc = cv.minMaxLoc(result)

        similarity = float(similarity)
        assert max_val >= similarity, f"Template not found ({max_val} < {similarity})"

        return (
            int(max_loc[0] + template.shape[1] / 2),
            int(max_loc[1] + template.shape[0] / 2),
        )

    def click_template(self, template, similarity=0.95):
        """Click center of the location best matching the given ``template`` if match with at least the given ``similarity`` threshold is found.

        Arguments:

        ``template``
            path to a image file used as the match template

        ``similarity``
            minimum accepted match similarity (default: 0.95).
        """
        assert os.path.isfile(template), f"File not found: {template}"
        x, y = self.match_template(template, similarity)
        Mouse.Instance.Click(Point(int(x), int(y)))
